# -*- coding: utf-8 -*-
"""
/***************************************************************************
OpenLayers Plugin
A QGIS plugin

                             -------------------
begin                : 2010-02-03
copyright            : (C) 2010 by Pirmin Kalberer, Sourcepole
email                : pka at sourcepole.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
import PyQt4
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4.QtWebKit import *
from PyQt4.QtNetwork import *
from qgis.core import *

from tools_network import getProxy


debuglevel = 4  # 0 (none) - 4 (all)


def debug(msg, verbosity=1):
    if debuglevel >= verbosity:
        qDebug(msg)


class OLWebPage(QWebPage):
    def __init__(self, parent=None):
        QWebPage.__init__(self, parent)
        self.__manager = None  # Need persist for PROXY
        # Set Proxy in webpage
        proxy = getProxy()
        if proxy is not None:
            self.__manager = QNetworkAccessManager()
            self.__manager.setProxy(proxy)
            self.setNetworkAccessManager(self.__manager)

        self.loaded = False

        self.extent = None
        self.olResolutions = None

        self.lastExtent = None
        self.lastViewPortSize = None
        self.lastLogicalDpi = None
        self.lastOutputDpi = None
        self.lastMapUnitsPerPixel = None

    def resolutions(self):
        if self.olResolutions is None:
            # get OpenLayers resolutions
            jsResolutions = self.mainFrame().evaluateJavaScript("map.layers[0].resolutions")
            debug("Detected OpenLayers resolutions: %s" % jsResolutions)
            self.olResolutions = jsResolutions
        return self.olResolutions or []

    def javaScriptConsoleMessage(self, message, lineNumber, sourceID):
        qDebug("%s[%d]: %s" % (sourceID, lineNumber, message))


class OpenlayersController(QObject):
    """
    Helper class that deals with QWebPage.
    The object lives in GUI thread, its request() slot is asynchronously called from worker thread.
    See https://github.com/wonder-sk/qgis-mtr-example-plugin for basic example
    1. Load Web Page with OpenLayers map
    2. Update OL map extend according to QGIS canvas extent
    """

    # signal that reports to the worker thread that the image is ready
    finished = pyqtSignal()

    def __init__(self, parent, context, webPage, layerType):
        QObject.__init__(self, parent)

        debug("OpenlayersController.__init__", 3)
        self.context = context
        self.layerType = layerType

        self.img = QImage()

        self.page = webPage
        self.page.loadFinished.connect(self.pageLoaded)

        self.timer = QTimer()
        self.timer.setInterval(100)
        self.timer.timeout.connect(self.checkMapUpdate)

        self.timerMax = QTimer()
        self.timerMax.setSingleShot(True)
        self.timerMax.setInterval(3000)  # TODO: different timeouts for map types
        self.timerMax.timeout.connect(self.mapTimeout)

    @pyqtSlot()
    def request(self):
        debug("[GUI THREAD] Processing request", 3)
        self.cancelled = False

        if not self.page.loaded:
            self.init_page()
        else:
            self.setup_map()

    def init_page(self):
        url = self.layerType.html_url()
        debug("page file: %s" % url)
        self.page.mainFrame().load(QUrl(url))
        if not self.layerType.emitsLoadEnd:
            debug("Layer type does not emit LoadEnd: connecting signal repaintRequested", 3)
            QObject.connect(self.page, SIGNAL("repaintRequested(QRect)"), self.pageRepaintRequested)
        # wait for page to finish loading
        debug("OpenlayersWorker waiting for page to load", 3)

    def pageLoaded(self):
        debug("[GUI THREAD] pageLoaded", 3)
        if self.cancelled:
            self.img.fill(Qt.gray)

        self.page.loaded = True

        if self.page.extent is None:
            self.setup_map()
        else:
            self.finished.emit()

    def setup_map(self):
        rendererContext = self.context

        self.outputDpi = rendererContext.painter().device().logicalDpiX()  # FIXME: self.mapSettings.outputDpi()
        debug(" extent: %s" % rendererContext.extent().toString(), 3)
        debug(" center: %lf, %lf" % (rendererContext.extent().center().x(), rendererContext.extent().center().y()), 3)
        debug(" size: %d, %d" % (rendererContext.painter().viewport().size().width(), rendererContext.painter().viewport().size().height()), 3)
        debug(" logicalDpiX: %d" % rendererContext.painter().device().logicalDpiX(), 3)
        debug(" outputDpi: %lf" % self.outputDpi)
        debug(" mapUnitsPerPixel: %f" % rendererContext.mapToPixel().mapUnitsPerPixel(), 3)
        #debug(" rasterScaleFactor: %s" % str(rendererContext.rasterScaleFactor()), 3)
        #debug(" outputSize: %d, %d" % (self.iface.mapCanvas().mapRenderer().outputSize().width(), self.iface.mapCanvas().mapRenderer().outputSize().height()), 3)
        #debug(" scale: %lf" % self.iface.mapCanvas().mapRenderer().scale(), 3)

        if (self.page.lastExtent != rendererContext.extent()
                or self.page.lastViewPortSize != rendererContext.painter().viewport().size()
                or self.page.lastLogicalDpi != rendererContext.painter().device().logicalDpiX()
                or self.page.lastOutputDpi != self.outputDpi
                or self.page.lastMapUnitsPerPixel != rendererContext.mapToPixel().mapUnitsPerPixel()):
            olSize = rendererContext.painter().viewport().size()
            if rendererContext.painter().device().logicalDpiX() != int(self.outputDpi):
                # use screen dpi for printing
                #sizeFact = self.outputDpi / 25.4 / rendererContext.mapToPixel().mapUnitsPerPixel()
                sizeFact = 1
                olSize.setWidth(rendererContext.extent().width() * sizeFact)
                olSize.setHeight(rendererContext.extent().height() * sizeFact)
            debug(" olSize: %d, %d" % (olSize.width(), olSize.height()), 3)
            self.page.setViewportSize(olSize)
            targetWidth = olSize.width()
            targetHeight = olSize.height()

            # find best resolution or use last
            qgisRes = rendererContext.extent().width() / targetWidth
            olRes = qgisRes
            for res in self.page.resolutions():
                olRes = res
                if qgisRes >= res:
                    break

            # adjust OpenLayers viewport to match QGIS extent
            olWidth = rendererContext.extent().width() / olRes
            olHeight = rendererContext.extent().height() / olRes
            debug("    adjust viewport: %f -> %f: %f x %f" % (qgisRes, olRes, olWidth, olHeight), 3)
            self.viewportSize = QSize(olWidth, olHeight)
            self.page.setViewportSize(self.viewportSize)
            self.img = QImage(self.viewportSize, QImage.Format_ARGB32_Premultiplied)

            if rendererContext.extent() != self.page.extent:
                self.page.extent = rendererContext.extent()  # FIXME: store seperate for each rendererContext
                debug("updating OpenLayers extent (%f, %f, %f, %f)" % (self.page.extent.xMinimum(), self.page.extent.yMinimum(), self.page.extent.xMaximum(), self.page.extent.yMaximum()), 3)
                self.page.mainFrame().evaluateJavaScript(
                    "map.zoomToExtent(new OpenLayers.Bounds(%f, %f, %f, %f), true);" %
                    (self.page.extent.xMinimum(), self.page.extent.yMinimum(), self.page.extent.xMaximum(), self.page.extent.yMaximum()))
                debug("map.zoomToExtent finished", 3)

            self.mapFinished = False
            self.timer.start()
            self.timerMax.start()

    def checkMapUpdate(self):
        if self.layerType.emitsLoadEnd:
            debug('waiting for loadEnd', 3)
            # wait for OpenLayers to finish loading
            # NOTE: does not work with Google and Yahoo layers as they do not emit loadstart and loadend events
            loadEndOL = self.page.mainFrame().evaluateJavaScript("loadEnd")
            debug('waiting ' + str(self.context.renderingStopped()) + ' ' + str(loadEndOL), 4)
            #debug("loadEndOL: %d" % loadEndOL, 3)
            if not loadEndOL is None:
                self.mapFinished = loadEndOL
            else:
                debug("OpenlayersLayer Warning: Could not get loadEnd")

        if self.mapFinished:
            self.timerMax.stop()
            self.timer.stop()

            self.renderMap()

            self.finished.emit()

    def pageRepaintRequested(self, rect):
        debug("OpenlayersLayer pageRepaintRequested", 3)
        #Plugin 1.1.2 finished rendering after waiting for 500ms timer

    def renderMap(self):
        rendererContext = self.context
        if rendererContext.painter().device().logicalDpiX() != int(self.outputDpi):
            printScale = 25.4 / self.outputDpi  # OL DPI to printer pixels
            rendererContext.painter().scale(printScale, printScale)

        # render OpenLayers to image
        painter = QPainter(self.img)
        self.page.mainFrame().render(painter)
        painter.end()

        # if olWidth != targetWidth or olHeight != targetHeight:
        #     # scale using QImage for better quality
        #     img = img.scaled(targetWidth, targetHeight, Qt.KeepAspectRatio, Qt.SmoothTransformation)
        #     debug("    scale image: %i x %i -> %i x %i" % (olWidth, olHeight, targetWidth, targetHeight), 3)

        # save current state
        self.page.lastExtent = rendererContext.extent()
        self.page.lastViewPortSize = rendererContext.painter().viewport().size()
        self.page.lastLogicalDpi = rendererContext.painter().device().logicalDpiX()
        self.page.lastOutputDpi = self.outputDpi
        self.page.lastMapUnitsPerPixel = rendererContext.mapToPixel().mapUnitsPerPixel()

    def mapTimeout(self):
        self.timer.stop()
        self.img.fill(Qt.gray)
        self.finished.emit()


class OpenlayersRenderer(QgsMapLayerRenderer):
    def __init__(self, layer, context, webPage, layerType):
        """ Initialize the object. This function is still run in the GUI thread.
            Should refrain from doing any heavy work.
        """
        QgsMapLayerRenderer.__init__(self, layer.id())
        self.context = context
        self.controller = OpenlayersController(None, context, webPage, layerType)
        self.loop = None

    def render(self):
        """ do the rendering. This function is called in the worker thread """

        debug("[WORKER THREAD] Calling request() asynchronously", 3)
        QMetaObject.invokeMethod(self.controller, "request")

        # setup a timer that checks whether the rendering has not been stopped in the meanwhile
        timer = QTimer()
        timer.setInterval(50)
        timer.timeout.connect(self.onTimeout)
        timer.start()

        debug("[WORKER THREAD] Waiting for the async request to complete", 3)
        self.loop = QEventLoop()
        self.controller.finished.connect(self.loop.exit)
        self.loop.exec_()

        debug("[WORKER THREAD] Async request finished", 3)

        painter = self.context.painter()
        painter.drawImage(0, 0, self.controller.img)
        return True

    def onTimeout(self):
        """ periodically check whether the rendering should not be stopped """
        if self.context.renderingStopped():
            debug("[WORKER THREAD] Cancelling rendering", 3)
            self.loop.exit()


class OpenlayersLayer(QgsPluginLayer):

    LAYER_TYPE = "openlayers"
    MAX_ZOOM_LEVEL = 15
    SCALE_ON_MAX_ZOOM = 13540  # QGIS scale for 72 dpi

    def __init__(self, iface, olLayerTypeRegistry):
        QgsPluginLayer.__init__(self, OpenlayersLayer.LAYER_TYPE, "OpenLayers plugin layer")
        self.setValid(True)

        self.olLayerTypeRegistry = olLayerTypeRegistry

        #self.iface = iface
        self.olWebPage = OLWebPage(self)

        #Set default layer type
        self.setLayerType(self.olLayerTypeRegistry.getById(0))

    def readXml(self, node):
        # custom properties
        self.setLayerType(self.olLayerTypeRegistry.getById(int(node.toElement().attribute("ol_layer_type", "0"))))
        return True

    def writeXml(self, node, doc):
        element = node.toElement()
        # write plugin layer type to project (essential to be read from project)
        element.setAttribute("type", "plugin")
        element.setAttribute("name", OpenlayersLayer.LAYER_TYPE)
        # custom properties
        element.setAttribute("ol_layer_type", str(self.layerType.id))
        return True

    def setLayerType(self, layerType):
        self.layerType = layerType
        coordRefSys = self.layerType.coordRefSys(None)  # FIXME
        self.setCrs(coordRefSys)
        #TODO: get extent from layer type
        self.setExtent(QgsRectangle(-20037508.34, -20037508.34, 20037508.34, 20037508.34))

    def createMapRenderer(self, context):
        return OpenlayersRenderer(self, context, self.olWebPage, self.layerType)
